/* vim:tabstop=4:expandtab:shiftwidth=4
 * 
 * Idesk -- XImlib2Caption.cpp
 *
 * Copyright (c) 2002, Chris (nikon) (nikon@sc.rr.com)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *      
 *      Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *      
 *      Neither the name of the <ORGANIZATION> nor the names of its
 *      contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * (See the included file COPYING / BSD )
 */
#include  <X11/extensions/shape.h>
#include "XImlib2Caption.h"
#include "XIconWithShadow.h"

XImlib2Caption::XImlib2Caption(AbstractContainer * c, AbstractIcon * iParent,
                   AbstractConfig * con, AbstractIconConfig * iConfig) 
           : AbstractCaption(c, iParent, con, iConfig),
             XImlib2Image(c, iParent, con, iConfig), FONTSHIFT(4)
{
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);

    visual = imlib_context_get_visual();
    cmap = imlib_context_get_colormap();

    configure();
    createFont();
}

XImlib2Caption::~XImlib2Caption()
{
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);

    XftColorFree( xContainer->getDisplay(), visual, cmap, &fontColor );

    if (shadowOn)
        XftColorFree( xContainer->getDisplay(), visual, cmap, &shadowFontColor);

     XFreePixmap(xContainer->getDisplay(), tmpPixmap);
     XFreeGC(xContainer->getDisplay(), gc);
     XFreeGC(xContainer->getDisplay(), bufferGc);

    XDestroyWindow(xContainer->getDisplay(), window);
}

void XImlib2Caption::configure()
{
    DesktopIconConfig * dIconConfig =
        dynamic_cast<DesktopIconConfig *>(AbstractImage::iconConfig);

    //grab the text
    text = AbstractImage::iconConfig->getCaption();
    
    //set boldness
    if (dIconConfig->getBoldness())
        boldVal = XFT_WEIGHT_BOLD;
    else
        boldVal = XFT_WEIGHT_MEDIUM;

    //set shadow options
    shadowOn = dIconConfig->isShadow();
    shadowX = dIconConfig->getShadowX();
    shadowY = dIconConfig->getShadowY();
    shadowColor = dIconConfig->getShadowColor();

    //captionOnHover
    captionOnHover = dIconConfig->getCaptionOnHover();
    
    //captionPlacement
    captionPlacement = dIconConfig->getCaptionPlacement();
}

void XImlib2Caption::createFont()
{
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);
    DesktopIconConfig * dIconConfig =
        dynamic_cast<DesktopIconConfig *>(AbstractImage::iconConfig);

    if (!xContainer) cout << "!xContainer\n";
    if (!dIconConfig) cout << "!dIconConfig\n";

    font = XftFontOpen( xContainer->getDisplay(),
                        DefaultScreen( xContainer->getDisplay()), 
                        XFT_FAMILY,
                        XftTypeString,
                        dIconConfig->getFont().c_str(),
                        XFT_SIZE,
                        XftTypeDouble,
                        (double)dIconConfig->getFontSize(), 
                        XFT_WEIGHT, XftTypeInteger, boldVal,
                        NULL );
    
    //cout << "xContainer->getDisplay()" << xContainer->getDisplay() << endl;
    //cout << "Color: " << dIconConfig->getFontColor() << endl;
    
    XColor  screen, exact;
    if (!XAllocNamedColor (xContainer->getDisplay(), cmap, (char*)dIconConfig->getFontColor().c_str(), &screen, &exact))
        cout << "BAD COLORS!!!\n";
    
    XftColorAllocName(xContainer->getDisplay(), visual, cmap,
                      (char*)dIconConfig->getFontColor().c_str(), &fontColor );

    if (shadowOn)
        XftColorAllocName(xContainer->getDisplay(), visual, cmap,
                          (char*)shadowColor.c_str(), &shadowFontColor );
}


Window * XImlib2Caption::getWindow()
{
    return &window;
}

void XImlib2Caption::createWindow()
{
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);

    XSetWindowAttributes attr;

    attr.override_redirect = True;
    DesktopConfig * dConfig =
		    dynamic_cast<DesktopConfig *>(XImlib2Image::config);
    /*Fixxxer*/
    attr.cursor = XCreateFontCursor(xContainer->getDisplay(), dConfig->getCursorOver());
    
    attr.event_mask = SubstructureRedirectMask |
                      SubstructureNotifyMask   |
                      ButtonPressMask          |
                      ButtonReleaseMask        |
                      PointerMotionMask        |
                      PropertyChangeMask       |
		      VisibilityChangeMask     |
                      ExposureMask;

    XftTextExtents8( xContainer->getDisplay(), font, (XftChar8*)text.c_str(), text.length(), &fontInfo );

    //cout << "Texto -->"<< text.c_str() << " Longitud ->" << text.length() << endl;
    
    // fix window extents so shadow text is not cut
    // TODO: fix window extents if shadow is negative:
    // Would have to edit x and y of the window
    width = fontInfo.width + FONTSHIFT;
    height = font->height;

    if (shadowOn)
    {
        width += shadowX > 0 ? shadowX : 0;
        height += shadowY > 0 ? shadowY : 0;
    }
    window = XCreateWindow( xContainer->getDisplay(),
                            xContainer->getRootWindow(),
                            0,
                            0,
                            width,
                            height,
                            0,
                            CopyFromParent,
                            CopyFromParent,
                            CopyFromParent,
			    CWOverrideRedirect|CWEventMask|CWCursor,
                            &attr );
    renderFontToImlib();
    //shapeWindow();
}

void XImlib2Caption::updateText()
{   
	XDesktopContainer * xContainer =
			dynamic_cast<XDesktopContainer *>(AbstractImage::container);

    //New Code 0.61
     XSetForeground(xContainer->getDisplay(), gc, 0xffffff);
     XSetLineAttributes(xContainer->getDisplay(), gc, 2, LineSolid, CapRound,
                                        JoinRound);
     XDrawArc(xContainer->getDisplay(), window, gc, 0, 3, width, height, 0, 0);

	if (shadowOn)
		XftDrawString8( fontDrawHandle, &shadowFontColor, font, FONTSHIFT/2 + shadowX, font->ascent + shadowY, (XftChar8*)text.c_str(), text.length() );

	XftDrawString8( fontDrawHandle, &fontColor, font, FONTSHIFT/2, font->ascent, (XftChar8*)text.c_str(), text.length() );
}


void XImlib2Caption::renderFontToImlib()
{
    // NEW CODE
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);

    tmpPixmap = XCreatePixmap(xContainer->getDisplay(), window, width,
                            height,
                            DefaultDepth(xContainer->getDisplay(), 
                            DefaultScreen(xContainer->getDisplay())));
    
    gc = XCreateGC(xContainer->getDisplay(), window, 0, 0);
    
    
    if( tmpPixmap == (Pixmap) NULL )
    {
	    printf("\nError creando Pixmap\n");
	    XCloseDisplay(xContainer->getDisplay());
	    exit(1);
    }
    
    //Fixed 0.61
    bufferGc = XCreateGC(xContainer->getDisplay(), tmpPixmap, 0, 0); 
    XSetForeground(xContainer->getDisplay(), bufferGc, BlackPixel(xContainer->getDisplay(), 0));
    XSetLineAttributes(xContainer->getDisplay(), bufferGc, 2, LineSolid,
                                   CapRound, JoinRound);
    
    fontDrawHandle = XftDrawCreate(xContainer->getDisplay(), tmpPixmap, visual,
				   cmap);

    //XSetBackground(xContainer->getDisplay(), gc, WhitePixel(xContainer->getDisplay(), 0));
    //updateText(); //draw text onto the buffer
    //XSetFillStyle(xContainer->getDisplay(), gc, FillSolid);
    //XSetTile(xContainer->getDisplay(), gc, tmpPixmap);
    //XFillRectangle(xContainer->getDisplay(), window, gc, 0, 0, width, height);

    //Pixmap shapeMask = XCreatePixmap(xContainer->getDisplay(), window, width, height,1);
    //GC tempGc = XCreateGC(xContainer->getDisplay(), shapeMask, 0, 0);
    
    //XCopyPlane(xContainer->getDisplay(), tmpPixmap, shapeMask, tempGc, 0, 0,
	 //      width, height, 0, 0, 1);
    //XShapeCombineMask(xContainer->getDisplay(), window, 0, 0, 0, shapeMask, ShapeSubtract);
  
    //imlib_context_set_drawable(tmpPixmap);
    
    //image = imlib_create_image_from_drawable(0, 0, 0, width, height, 0);

    //XFreePixmap(xContainer->getDisplay(), tmpPixmap);    
}

void XImlib2Caption::draw()
{
   //apply the buffer to the window 
	//cout << "XImlib2Caption::draw--->\n";
	XDesktopContainer * xContainer =
			dynamic_cast<XDesktopContainer *>(AbstractImage::container);
	XIconWithShadow * sIcon = dynamic_cast<XIconWithShadow *>(AbstractImage::iconParent);
    
    //ImlibImage *textBackground;
	int fX, fY;

    // using the spareRoot image (the background image created at startup) 
    // instead of dynamically geting the window to speed things up.
    
	setXY(fX, fY);

	imlib_context_set_image(*xContainer->getSpareRoot());
	imlib_context_set_drawable(tmpPixmap);

        //this doesn't work for some reason, bug?
        imlib_render_image_part_on_drawable_at_size(fX, fY, width , height ,
                               0, 0, width , height );

	Imlib_Image cropImage = imlib_create_cropped_image(fX, fY, width, height);
	imlib_context_set_image(cropImage);
	imlib_render_image_on_drawable(0, 0);
	imlib_free_image();
	imlib_context_set_image(*xContainer->getSpareRoot());
    
    // if the icon shadow is enabled and the shadow is visible underneath the
    // text window
	if (sIcon && //TODO Fix this
		   sIcon->getShadowX() > fX - sIcon->getWidth() &&
		   sIcon->getShadowY() > fY - sIcon->getHeight() &&
		   sIcon->getShadowX() < fX + sIcon->getWidth() &&
		   sIcon->getShadowY() < fY + sIcon->getHeight() )
		sIcon->renderShadowToImage(tmpPixmap, fX, fY);

	updateText();

	XMoveWindow( xContainer->getDisplay(), window, fX, fY );
    
    //apply the buffer to the window
	XSetFillStyle(xContainer->getDisplay(), gc, FillTiled);
	XSetTile(xContainer->getDisplay(), gc, tmpPixmap);
	XFillRectangle(xContainer->getDisplay(), window, gc, 0, 0, width, height);
}

void XImlib2Caption::lowerWindow()
{   
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);

    XLowerWindow( xContainer->getDisplay(), window );
}

void XImlib2Caption::setXY(int & fX, int & fY)
{
    //TODO fix pixel placement
    if (captionPlacement == "Right")
        fX = AbstractImage::iconParent->getX() + AbstractImage::iconParent->getWidth() + 10;
    else if (captionPlacement == "Left")
        fX = AbstractImage::iconParent->getX() - fontInfo.width - 10;
    else //(captionPlacement == "Bottom" || captionPlacement == "Top")
        fX = AbstractImage::iconParent->getX() + 
             (AbstractImage::iconParent->getWidth() - fontInfo.width)/2;

    if (captionPlacement == "Left" || captionPlacement == "Right")
        fY = AbstractImage::iconParent->getY() + (AbstractImage::iconParent->getHeight() - fontInfo.height)/2;
    else if (captionPlacement == "Top")
        fY = AbstractImage::iconParent->getY() - fontInfo.height - 2*FONTSHIFT;
    else //(captionPlacement == "Bottom")
        fY = AbstractImage::iconParent->getY() + AbstractImage::iconParent->getHeight() + FONTSHIFT;
}

int XImlib2Caption::getFontWidth()
{
    return fontInfo.width;
}

int XImlib2Caption::getFontHeight()
{
    return font->height;
}
void XImlib2Caption::mapWindow()
{
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);

    XMapWindow( xContainer->getDisplay(), window );
}

void XImlib2Caption::initMapWindow()
{
    if (!captionOnHover)
        mapWindow();
}

void XImlib2Caption::unmapWindow()
{
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);

    XUnmapWindow( xContainer->getDisplay(), window );
}

void XImlib2Caption::moveWindow(int xC, int yC)
{
    XDesktopContainer * xContainer =
        dynamic_cast<XDesktopContainer *>(AbstractImage::container);
    //int xC, yC;
    setXY(xC, yC);
    XMoveWindow( xContainer->getDisplay(), window, xC, yC );
}

void XImlib2Caption::mouseOverEffect()
{
    if (captionOnHover)
        mapWindow();
}

void XImlib2Caption::mouseOffEffect()
{
    if (captionOnHover)
        unmapWindow();
}
